/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL
v2. You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"
#include "defs.h"
#include "system/sm_meta.h"
#include <cstddef>

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query
 */
std::shared_ptr<Query>
Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse) {
  // 接受的参数为一个抽象语法树
  std::shared_ptr<Query> query = std::make_shared<Query>();

  if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse)) {
    // 处理表名（这些表名是在form子句中使用的表名）
    query->tables = std::move(x->tabs);
    /** DONE by noah: 检查表是否存在 */
    for (auto &tab_name : query->tables) {
      if (!sm_manager_->db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
      }
    }
    // 处理target list，再target list中添加上表名，例如 a.id
    for (auto &sv_sel_col : x->cols) {
      TabCol sel_col = {.tab_name = sv_sel_col->tab_name,
                        .col_name = sv_sel_col->col_name};
      query->cols.push_back(sel_col);
    }

    std::vector<ColMeta> all_cols;
    get_all_cols(query->tables, all_cols);
    if (query->cols.empty()) {
      // select all columns
      for (auto &col : all_cols) {
        TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name};
        query->cols.push_back(sel_col);
      }
    } else {
      // infer table name from column name
      for (auto &sel_col : query->cols) {
        sel_col = check_column(all_cols, sel_col); // 列元数据校验
      }
    }
    // 处理where条件
    get_clause(x->conds, query->conds);
    check_clause(query->tables, query->conds);
  } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
    /** DONE by noah: */
    // 进行语义检查，检查表是否存在
    if (!sm_manager_->db_.is_table(x->tab_name)) {
      throw TableNotFoundError(x->tab_name);
    }
    // 处理set子句。在AST中所有的SET节点都没有表名，所以这里需要遍历所有的set，为set加上表名后加入到query->set_clauses中
    for (auto &sv_set : x->set_clauses) {
      TabCol set_col = {.tab_name = x->tab_name, .col_name = sv_set->col_name};
      //   fix by liam:
      //   在这里调整类型，进行语义分析，对于传入的int，需要float时进行合理转换
      //   先取出table的col元数据
      TabMeta &tab = sm_manager_->db_.get_table(x->tab_name);
      std::vector<ColMeta>::iterator col = tab.get_col(sv_set->col_name);
      query->set_clauses.push_back(
          {set_col, convert_sv_value(sv_set->val, *col)});
    }
    // 处理where条件
    get_clause(x->conds, query->conds);
    check_clause({x->tab_name}, query->conds);
  } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
    // 处理where条件
    get_clause(x->conds, query->conds);
    check_clause({x->tab_name}, query->conds);
  } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
    // done by noah：处理insert的类型转换
    // TO BE COMPLETED ... liamY
    std::vector<ColMeta> all_cols;
    // // 获取所有列（在insert的时候只会有一张表，所以可以只给一张表的名字）
    get_all_cols({x->tab_name}, all_cols);
    // // 一次性检查所有value的类型
    // check_insert_type(all_cols,x->vals);
    // 处理insert 的values值
    for (size_t i = 0; i < all_cols.size(); i++) {
      std::shared_ptr<ast::Value> &sv_val = x->vals[i];
      // 经过类型转换之后再将value压入查询
      query->values.push_back(convert_sv_value(sv_val, all_cols[i]));
    }
  } else {
    // do nothing
  }
  query->parse = std::move(parse);
  return query;
}

TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols,
                             TabCol target) {
  if (target.tab_name.empty()) {
    // Table name not specified, infer table name from column name
    std::string tab_name;
    for (auto &col : all_cols) {
      if (col.name == target.col_name) {
        if (!tab_name.empty()) {
          throw AmbiguousColumnError(target.col_name);
        }
        tab_name = col.tab_name;
      }
    }
    if (tab_name.empty()) {
      throw ColumnNotFoundError(target.col_name);
    }
    target.tab_name = tab_name;
  } else {
    /** DONE by noah: Make sure target column exists */
    // 获取表元数据
    TabMeta &tab = sm_manager_->db_.get_table(target.tab_name);
    if (!tab.is_col(target.col_name)) {
      // 不存在当前列的时候需要抛出异常
      throw ColumnNotFoundError(target.col_name);
    }
  }
  return target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names,
                           std::vector<ColMeta> &all_cols) {
  for (auto &sel_tab_name : tab_names) {
    // 这里db_不能写成get_db(), 注意要传指针
    const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
    all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
  }
}

void Analyze::get_clause(
    const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds,
    std::vector<Condition> &conds) {
  conds.clear();
  for (auto &expr : sv_conds) {
    Condition cond;
    cond.lhs_col = {.tab_name = expr->lhs->tab_name,
                    .col_name = expr->lhs->col_name};
    cond.op = convert_sv_comp_op(expr->op);
    if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
      cond.is_rhs_val = true;
      cond.rhs_val = convert_sv_value(rhs_val);
    } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
      cond.is_rhs_val = false;
      cond.rhs_col = {.tab_name = rhs_col->tab_name,
                      .col_name = rhs_col->col_name};
    }
    conds.push_back(cond);
  }
}

// 本函数只会检查where的条件是否符合语义规定（也会在这里进行类型转换）
void Analyze::check_clause(const std::vector<std::string> &tab_names,
                           std::vector<Condition> &conds) {
  // auto all_cols = get_all_cols(tab_names);
  std::vector<ColMeta> all_cols;
  get_all_cols(tab_names, all_cols);
  // Get raw values in where clause
  for (auto &cond : conds) {
    // Infer table name from column name
    cond.lhs_col = check_column(all_cols, cond.lhs_col);
    if (!cond.is_rhs_val) {
      cond.rhs_col = check_column(all_cols, cond.rhs_col);
    }
    TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
    auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
    ColType lhs_type = lhs_col->type;
    ColType rhs_type;
    if (cond.is_rhs_val) {
      // done by noah：如果类型不同尝试进行类型转换
      if (cond.rhs_val.type != lhs_col->type) {
        // 尝试修改类型
        // 判断类型是否能进行转换
        if (can_convert_type(cond.rhs_val.type, lhs_col->type)) {
          // 表示能进行类型转换，进行类型转换
          cond.rhs_val.set_type(lhs_col->type);
        }
        // else{
        //     throw IncompatibleTypeError(coltype2str(cond.rhs_val.type),
        //     coltype2str(lhs_col->type));
        // }
      }
      // 一定要在这一行之前进行类型转换，否则会导致init_raw函数报错（如果左右类型长度不同的话）
      // 经过类型转换之后左右的类型一定相同，可以对raw指针进行初始化
      cond.rhs_val.init_raw(lhs_col->len);
      rhs_type = cond.rhs_val.type;
    } else {
      TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
      auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
      rhs_type = rhs_col->type;
      // todo：如果列类型不同的话也要考虑类型转换
      // done by liamY;这里不能直接转换，不然有很多问题，直接在判断类型不符上动即可
      if (rhs_type != lhs_type) {
        if (can_convert_type(rhs_type, lhs_type)) {
          // 表示能进行类型转换，进行类型转换,这里把他们都转为float（因为目前只涉及）
        }
      }
    }
    // added by liamY:多一个容许类型转换的判断
    if (lhs_type != rhs_type && !can_convert_type(lhs_type, rhs_type)) {
      throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
    }
  }
}

// done by noah:检查insert类型函数
void Analyze::check_insert_type(std::vector<ColMeta> &all_cols,
                                std::vector<std::shared_ptr<Value>> &values) {
  if (all_cols.size() != values.size()) {
    throw InvalidValueCountError();
  }
  for (size_t i = 0; i < values.size(); i++) {
    auto &col = all_cols[i];
    auto &val = values[i];
    if (col.type != val->type) {
      throw IncompatibleTypeError(coltype2str(col.type),
                                  coltype2str(val->type));
    }
    val->init_raw(col.len);
    memcpy(val->raw->data, val->raw->data, col.len);
  }
}
// fix by liamY:加入colMeta参数用于增加数据的兼容性
Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val,
                                const ColMeta col) {
  Value val;
  if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
    // 当传入的参数是int值，但是对应的col是float类型的时候，需要进行类型转换
    if (col.type == TYPE_FLOAT) {
      val.set_float(int_lit->val);
    } else {
      val.set_int(int_lit->val);
    }
  } else if (auto float_lit =
                 std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
    // fix by liamY: 当传入的是float值，但是对应的col是int类型的时候，需要进行类型转换
    if (col.type == TYPE_INT) {
      val.set_int(float_lit->val);
    } else {
      val.set_float(float_lit->val);
    }
  } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
    val.set_str(str_lit->val);
  } else {
    throw InternalError("Unexpected sv value type");
  }
  return val;
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
  Value val;
  if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
    val.set_int(int_lit->val);
  } else if (auto float_lit =
                 std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
    val.set_float(float_lit->val);
  } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
    val.set_str(str_lit->val);
  } else {
    throw InternalError("Unexpected sv value type");
  }
  return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
  std::map<ast::SvCompOp, CompOp> m = {
      {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
      {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
  };
  return m.at(op);
}

// done by
// noah：can_convert_type函数，用于判断两个类型是否能进行类型转换（目前只考虑浮点型和整形之间的转换）
bool Analyze::can_convert_type(ColType from, ColType to) {
  // 如果两个类型相同，那么可以进行类型转换
  if (from == to) {
    return true;
  }
  // 如果from是int类型，那么可以转换为float类型
  if (from == TYPE_INT && to == TYPE_FLOAT) {
    return true;
  }
  // 如果from是float类型，那么可以转换为int类型
  if (from == TYPE_FLOAT && to == TYPE_INT) {
    return true;
  }
  return false;
}