/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL
v2. You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "defs.h"
#include "record/rm_defs.h"
#include "system/sm_meta.h"
#include <cassert>
#include <cstring>
#include <memory>
#include <string>
#include <vector>

struct TabCol {
  std::string tab_name;
  std::string col_name;

  friend bool operator<(const TabCol &x, const TabCol &y) {
    return std::make_pair(x.tab_name, x.col_name) <
           std::make_pair(y.tab_name, y.col_name);
  }
};

struct Value {
  ColType type; // type of value
  union {
    int int_val;     // int value
    float float_val; // float value
  };
  std::string str_val; // string value

  std::shared_ptr<RmRecord> raw; // raw record buffer

  void set_int(int int_val_) {
    type = TYPE_INT;
    int_val = int_val_;
  }

  void set_float(float float_val_) {
    type = TYPE_FLOAT;
    float_val = float_val_;
  }

  void set_str(std::string str_val_) {
    type = TYPE_STRING;
    str_val = std::move(str_val_);
  }

  // done by noah：类型设置函数,用于根据传入的类型设置Value的值
  void set_type(ColType type_) {
    // 根据原有的类型以及传入的类型修改联合的值
    if (type == TYPE_INT) {
      if (type_ == TYPE_FLOAT) {
        float_val = int_val;
      } else if (type_ == TYPE_STRING) {
        str_val = std::to_string(int_val);
      }
    } else if (type == TYPE_FLOAT) {
      if (type_ == TYPE_INT) {
        int_val = float_val;
      } else if (type_ == TYPE_STRING) {
        str_val = std::to_string(float_val);
      }
    } else if (type == TYPE_STRING) {
      if (type_ == TYPE_INT) {
        int_val = std::stoi(str_val);
      } else if (type_ == TYPE_FLOAT) {
        float_val = std::stof(str_val);
      }
    }
    // 设置类型
    type = type_;
  }

  void init_raw(int len) {
    assert(raw == nullptr);
    raw = std::make_shared<RmRecord>(len);
    if (type == TYPE_INT) {
      assert(len == sizeof(int));
      *(int *)(raw->data) = int_val;
    } else if (type == TYPE_FLOAT) {
      assert(len == sizeof(float));
      *(float *)(raw->data) = float_val;
    } else if (type == TYPE_STRING) {
      if (len < (int)str_val.size()) {
        throw StringOverflowError();
      }
      memset(raw->data, 0, len);
      memcpy(raw->data, str_val.c_str(), str_val.size());
    }
  }
};

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE };

struct Condition {
  TabCol lhs_col;  // left-hand side column
  CompOp op;       // comparison operator
  bool is_rhs_val; // true if right-hand side is a value (not a column)
  TabCol rhs_col;  // right-hand side column
  Value rhs_val;   // right-hand side value
  // by liamY:添加一个辅助函数，用于判断是否满足条件
  bool judge(const char *data, const std::vector<ColMeta> &cols) {
    // 先从cols中找到lhs_col和rhs_col对应的列
    const ColMeta *lhs_meta = nullptr;
    const ColMeta *rhs_meta = nullptr;
    Value lhs_val;
    const char *rhs_val_col;

    for (auto &col : cols) {
      if (lhs_col.col_name == col.name && lhs_col.tab_name == col.tab_name) {
        const char *lhs_val_col;
        lhs_val_col = data + col.offset;
        lhs_meta = &col;
        // 设置更新lhs_val值
        lhs_val.type = col.type;
        if (col.type == TYPE_INT) {
          lhs_val.int_val = *(int *)lhs_val_col;
        } else if (col.type == TYPE_FLOAT) {
          lhs_val.float_val = *(float *)lhs_val_col;
        } else if (col.type == TYPE_STRING) {
          lhs_val.str_val = std::string(lhs_val_col, col.len);
        }
      }
      if (!is_rhs_val && rhs_col.col_name == col.name &&
          rhs_col.tab_name == col.tab_name) {
        rhs_val_col = data + col.offset;
        // 设置更新rhs_val值
        rhs_val.type = col.type;
        if (col.type == TYPE_INT) {
          rhs_val.int_val = *(int *)rhs_val_col;
        } else if (col.type == TYPE_FLOAT) {
          rhs_val.float_val = *(float *)rhs_val_col;
        } else if (col.type == TYPE_STRING) {
          rhs_val.str_val = std::string(rhs_val_col, col.len);
        }
        rhs_meta = &col;
      }
    }
    if (lhs_meta == nullptr || (!is_rhs_val && rhs_meta == nullptr)) {
      // 说明没找到对应的列
      return false;
    }
    // 先根据op来分类操作
    switch (op) {
    case OP_EQ: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val == rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val == rhs_val.float_val;
        }
        // float 和 int的兼容情况
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val == rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val == rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) == 0;
        }
      }
    case OP_NE: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val != rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val != rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val != rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val != rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) != 0;
        }
    }
    case OP_LT: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val < rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val < rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val < rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val < rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) < 0;
        }
    }
    case OP_GT: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val > rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val > rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val > rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val > rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) > 0;
        }
      }
    case OP_LE: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val <= rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val <= rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val <= rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val <= rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) <= 0;
        }
      }
    case OP_GE: {
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val >= rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val >= rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_FLOAT && lhs_val.type == TYPE_INT) {
          return lhs_val.int_val >= rhs_val.float_val;
        }
        if (rhs_val.type == TYPE_INT && lhs_val.type == TYPE_FLOAT) {
          return lhs_val.float_val >= rhs_val.int_val;
        }
        if (rhs_val.type == TYPE_STRING) {
          return strcmp(lhs_val.str_val.c_str(), rhs_val.str_val.c_str()) >= 0;
        }
    }
    default:
      return true;
    }
  }
};

struct SetClause {
  TabCol lhs;
  Value rhs;
};