/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "system/sm_meta.h"
#include <vector>

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        std::vector<ColMeta> right_cols = right_->cols();
        // 注意这里没有修改原来right_cols的offset,comment added by liamY
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);

    }

    // done by liamY:重写父类的cols方法
    const std::vector<ColMeta> &cols() const override { return cols_; }
    // done by liamY: 写beginTuple初始化数据集
    void beginTuple() override {
        left_->beginTuple();
        right_->beginTuple();   
    }
    // done by liamY: 写nextTuple实现，移动数据集到满足条件的位置
    void nextTuple() override {
        while (!isend) {
            left_->nextTuple();
            if (left_->is_end()) {
                left_->beginTuple();
                right_->nextTuple();
                if (right_->is_end()) {
                    isend = true;
                    return;
                }
            }
            // 检验是否满足条件,通过next取出记录
            std::unique_ptr<RmRecord> rec = this->Next();
            // 通过fed_conds_判断是否满足条件
            bool flag = true;
            for (auto &cond : fed_conds_) {
                if(!cond.judge(rec->data, cols_)){
                    flag = false;
                    break;
                }
            }
            // 如果满足条件，就直接返回
            if (flag) {
                return;
            }
        }
    }
    // done by liamY: 重写Next，返回一条记录
    std::unique_ptr<RmRecord> Next() override {
        if (isend) {
            return nullptr;
        }
        auto left_rec = left_->Next();
        auto right_rec = right_->Next();
        auto new_rec = std::make_unique<RmRecord>(len_);
        for (auto &col: cols_){
            // 如果是左表的字段
            if (col.offset < left_->tupleLen()){
                memcpy(new_rec->data + col.offset, left_rec->data + col.offset, col.len);
            } else {
                memcpy(new_rec->data + col.offset, right_rec->data + col.offset - left_->tupleLen(), col.len);
            }
        }
        return new_rec;
    }

    // done by liamY:重写is_end方法
    bool is_end() const override { return isend; }

    // done by limaY:重写tupleLen方法
    size_t tupleLen() const override { return len_; }

    Rid &rid() override { return _abstract_rid; }
};