/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;        // 投影节点的儿子节点
    std::vector<ColMeta> cols_;                     // 需要投影的字段
    size_t len_;                                    // 字段总长度
    std::vector<size_t> sel_idxs_;                  // 就是每一个字段在原表中的位置（是第几个字段）

   public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &sel_cols) {
        prev_ = std::move(prev);

        size_t curr_offset = 0;
        auto &prev_cols = prev_->cols();
        for (auto &sel_col : sel_cols) {
            auto pos = get_col(prev_cols, sel_col);
            sel_idxs_.push_back(pos - prev_cols.begin());
            auto col = *pos;
            col.offset = curr_offset;
            curr_offset += col.len;
            cols_.push_back(col);
        }
        len_ = curr_offset;
    }

    // done by noah：重写父类的方法
    void beginTuple() override {
        // 采用递归向下分析来遍历算子树（也就是需要先执行底层的算子）
        prev_->beginTuple();
    }

    // done by noah：重写父类的方法
    void nextTuple() override {
        // 重写父类的nextTuple方法
        prev_->nextTuple();
    }

    // done by noah：重写父类的cols方法
    const std::vector<ColMeta> &cols() const override { return cols_; }

    // done by noah：重写父类的is_end方法
    bool is_end() const override { return prev_->is_end(); }

    // done by noah：重写父类的方法
    std::unique_ptr<RmRecord> Next() override {
        auto rec = prev_->Next();
        if (rec == nullptr) {
            return nullptr;
        }
        auto new_rec = std::make_unique<RmRecord>(len_);
        // 这一段代码就是进行投影操作，从原表中选取出指定的字段复制到目标new_rec中
        for (size_t i = 0; i < sel_idxs_.size(); i++) {
            auto &col = cols_[i];
            auto &sel_idx = sel_idxs_[i];
            memcpy(new_rec->data + col.offset, rec->data + prev_->cols()[sel_idx].offset, col.len);
        }
        return new_rec;
    }

    Rid &rid() override { return _abstract_rid; }
};