/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL
v2. You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "common/config.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
private:
  std::string tab_name_;             // 表的名称
  std::vector<Condition> conds_;     // scan的条件
  RmFileHandle *fh_;                 // 表的数据文件句柄
  std::vector<ColMeta> cols_;        // scan后生成的记录的字段
  size_t len_;                       // scan后生成的每条记录的长度
  std::vector<Condition> fed_conds_; // 同conds_，两个字段相同

  Rid rid_;
  std::unique_ptr<RecScan> scan_; // table_iterator

  SmManager *sm_manager_;

public:
  SeqScanExecutor(SmManager *sm_manager, std::string tab_name,
                  std::vector<Condition> conds, Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = std::move(tab_name);
    conds_ = std::move(conds);
    TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
    fh_ = sm_manager_->fhs_.at(tab_name_).get();
    cols_ = tab.cols;
    len_ = cols_.back().offset + cols_.back().len;

    context_ = context;

    fed_conds_ = conds_;
  }

  // done by noah：重写cols函数
  const std::vector<ColMeta> &cols() const override { return cols_; };

  // done by noah：重写is_end函数。实际上就是判断是否检索到文件末尾
  bool is_end() const override { return scan_->is_end(); }

  // done by noah：重写父类函数
  void beginTuple() override {
    // 重写父类的beginTuple函数
    // 1. 生成一个RecordScan对象
    scan_ = std::make_unique<RmScan>(fh_);
    // fix by liamY:
    // 2. 从RecordScan对象中获取第一个满足条件的Record
    rid_ = scan_->rid();
    // 如果rid里面的page_id为0,说明表是空的，这个时候不需要任何操作，直接返回
    if (rid_.page_no == 0 || rid_.slot_no == INVALID_PAGE_ID) {
      return;
    }
    // 判断是否满足条件
    bool flag = true;
    auto rec = scan_->rid();
    auto rm_rec = fh_->get_record(rec, context_);
    for (auto &cond : fed_conds_) {
      if (!cond.judge(rm_rec->data, cols_)) {
        rid_ = rec;
        flag = false;
        break;
      }
    }
    if (!flag) {
      nextTuple();
    }
  }

  // done by noah：重写父类函数
  // fix by liamY:
  void nextTuple() override {
    // 重写父类的nextTuple函数
    // 1. 从RecordScan对象中获取下一个Record
    // 注意这里要根据conds_来筛选到满足条件的下一个Record
    scan_->next();
    while (!scan_->is_end()) {
      auto rec = scan_->rid();
      auto rm_rec = fh_->get_record(rec, context_);
      // 判断是否满足条件
      bool flag = true;
      for (auto &cond : fed_conds_) {
        if (!cond.judge(rm_rec->data, cols_)) {
          flag = false;
          break;
        }
      }
      if (flag) {
        rid_ = rec;
        return;
      }
      scan_->next();
    }
  }
  // fix by liamY
  std::unique_ptr<RmRecord> Next() override {
    // 根据rid获取RmRecord
    auto rm_rec = fh_->get_record(rid_, context_);
    return rm_rec;
  }

  // done by liamY: 重写父类的tupleLen()函数
  size_t tupleLen() const override { return len_; }

  Rid &rid() override { return rid_; }
};