/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL
v2. You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include <cstdio>

class UpdateExecutor : public AbstractExecutor {
private:
  TabMeta tab_;
  std::vector<Condition> conds_;
  RmFileHandle *fh_;
  std::vector<Rid> rids_;
  std::string tab_name_;
  std::vector<SetClause> set_clauses_;
  SmManager *sm_manager_;

public:
  UpdateExecutor(SmManager *sm_manager, const std::string &tab_name,
                 std::vector<SetClause> set_clauses,
                 std::vector<Condition> conds, std::vector<Rid> rids,
                 Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = tab_name;
    set_clauses_ = set_clauses;
    tab_ = sm_manager_->db_.get_table(tab_name);
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    conds_ = conds;
    rids_ = rids;
    context_ = context;
  }

  // done by noah：完成更新操作
  // continued by liamY
  std::unique_ptr<RmRecord> Next() override {
    // 此时rids_中存储的是需要更新的记录的位置，只需要相应的更新即可
    // TODO:liamY 优化的时候这里的遍历可以调整，可能用一个数组结构存储信息更好
    for (auto &rid : rids_) {
      // 根据setClause 挨个设置每个rid的值
      for (auto set_clause : set_clauses_) {
        // set_clause 里面指明了table 和 col信息以及应该设置的值
        std::vector<ColMeta>::iterator col = tab_.cols.end();
        if (set_clause.lhs.tab_name == tab_name_) {
          col = tab_.get_col(set_clause.lhs.col_name);
        }
        // 如果col为空，说明没有找到对应的列
        if (col == tab_.cols.end()) {
          throw ColumnNotFoundError(set_clause.lhs.col_name);
        }
        // 如果col不为空，说明找到了对应的列,我们就更新该rid的对应的列的值
        // 先取出原纪录，改变值，再写回
        auto rm_rec = fh_->get_record(rid, context_);
        // 根据set_clause的值设置rm_rec的值
        if (col->type == TYPE_INT) {
          int *val = (int *) (rm_rec->data+col->offset);
          *val = set_clause.rhs.int_val;
        } else if (col->type == TYPE_FLOAT) {
          float *val = (float *) (rm_rec->data+col->offset);
          *val = set_clause.rhs.float_val;
        } else if (col->type == TYPE_STRING) {
          // 字符串类型的处理
          // 如果字符串长度大于列的长度，抛出异常
          if (set_clause.rhs.str_val.size() > col->len) {
            throw StringTooLongError(set_clause.rhs.str_val);
          }
          // 否则，将字符串拷贝到rm_rec中
          memcpy(rm_rec->data + col->offset, set_clause.rhs.str_val.c_str(),
                 set_clause.rhs.str_val.size());
          // 将剩余的位置填充为0
          memset(rm_rec->data + col->offset + set_clause.rhs.str_val.size(), 0,
                 col->len - set_clause.rhs.str_val.size());
        }
        // 更新记录
        fh_->update_record(rid, rm_rec->data, context_);
      }
    }
    printf("update record success\n");
    // 更新结束之后返回空指针以满足函数签名要求
    return nullptr;
  }

  Rid &rid() override { return _abstract_rid; }
};