/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"
#include "system/sm_manager.h"
#include "record/rm_manager.h"
std::unordered_map<txn_id_t, Transaction *> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction * TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
    // done by noah:
    // 1. 判断传入事务参数是否为空指针
    // 2. 如果为空指针，创建新事务
    // 3. 把开始事务加入到全局事务表中
    // 4. 返回当前事务指针
    // 判断当前事务是否为空指针
    if (txn == nullptr) {
        // 创建新事务
        // 获取新事务id并更新next_txn_id_
        txn_id_t txn_id = next_txn_id_.fetch_add(1);
        // 使用默认的隔离级别，所以只需要传入事务id即可
        Transaction myTransaction(txn_id);
        txn= &myTransaction;
    }
    // 事务begin时都需要将当前事务加入到全局事务表中
    // 获取txn_map的锁
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务加入到全局事务表中
    txn_map[txn->get_transaction_id()] = txn;
    // 释放锁
    lock.unlock();
    // todo:记录日志 
    return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // 目前还没有涉及到事务锁相关的部分，所以这里暂时先不管lock_set。后续需要进行并发控制时再进行相关锁的申请
    // Todo by noah:
    // 1. 如果存在未提交的写操作，提交所有的写操作
    // 2. 释放所有锁
    // 3. 释放事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    // 检查是否有未提交的写操作
    if (txn->get_write_set()->size() > 0) {
        // 提交所有的写操作
        for (auto write : *(txn->get_write_set())) {
            // 获取RmManager
            RmManager* rm_manager = sm_manager_->get_rm_manager();
            // 根据write中的表名获取文件句柄
            std::unique_ptr<RmFileHandle> file_hdr = rm_manager->open_file(write->GetTableName());
            // 向指定位置写入record。暂时还没办法根据事务拿到context，所以暂时先传入nullptr
            file_hdr->update_record(write->GetRid(), write->GetRecord().data, nullptr);
            // 关闭文件句柄
            rm_manager->close_file(file_hdr.get());
        }
    }
    // 释放所有锁
    for (auto lock : *(txn->get_lock_set())) {
        lock_manager_->unlock(txn,lock);
    }
    // 释放事务相关资源
    txn->get_lock_set()->clear();
    txn->get_write_set()->clear();
    // 把事务日志刷入磁盘中
    log_manager->flush_log_to_disk();
    // 更新事务状态
    txn->set_state(TransactionState::COMMITTED);
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction * txn, LogManager *log_manager) {
    // Todo:
    // 1. 回滚所有写操作
    // 2. 释放所有锁
    // 3. 清空事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    
}